/*****************************************************************
/    This file is part of PC/SC Lite driver for SM1.
/
/    PC/SC Lite driver for SM1 is free software; you can redistribute it and/or
/    modify it under the terms of the GNU General Public License as published by
/    the Free Software Foundation; either version 2 of the License, or
/    (at your option) any later version.
/
/    PC/SC Lite driver for SM1 is distributed in the hope that it will be
/    useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
/    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/    GNU General Public License for more details.
/
/    You should have received a copy of the GNU General Public License
/    along with PC/SC Lite driver for SM1; if not, write to the Free Software
/    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/
/    Copyright Damien Sauveron, 2003
/    contact : damien.sauveron@labri.fr
/
/ File   :   sm.c
/ Author :   Damien Sauveron <damien.sauveron@labri.fr>
/ Date   :   July 15, 2003
/ Purpose:   This provides reader specific low-level calls.
/            See http://www.linuxnet.com for more information.
/ License:   See file COPYING
/
******************************************************************/

#include "sm.h"
#include <unistd.h>


static struct sc_desc_s *Readers[PCSCLITE_MAX_CHANNELS];

// Check if the Lun is not to large for the ReadersSlots table
// returns TRUE in case of error
int iLunCheck(DWORD Lun)
{
  if ((LunToReaderIndex(Lun) >= PCSCLITE_MAX_CHANNELS) ||
      LunToReaderIndex(Lun) < 0)
    return TRUE;

  return FALSE;
} /* iLunCheck */

RESPONSECODE Open(DWORD Lun, DWORD Channel) {
  struct sc_desc_s *sc;
  char dev_name[FILENAME_MAX];
  sprintf(dev_name, "/dev/pcsc/%X", Channel);

  sc = smartmouse_open(dev_name);
  if (sc == NULL)
    return IFD_COMMUNICATION_ERROR;
   
  Readers[LunToReaderIndex(Lun)] = sc;

  return IFD_SUCCESS;
}

RESPONSECODE Close(DWORD Lun) {
  struct sc_desc_s *sc = Readers[LunToReaderIndex(Lun)];

  smartmouse_down(sc);
  smartmouse_close(sc);

  return IFD_SUCCESS;
}

RESPONSECODE IsPresent(DWORD Lun) {

  if (smartmouse_ispresent(Readers[LunToReaderIndex(Lun)]) == 0)
    return IFD_ICC_NOT_PRESENT;
  else
    return IFD_ICC_PRESENT;
}

RESPONSECODE PowerDown(DWORD Lun) {
  struct sc_desc_s *sc = Readers[LunToReaderIndex(Lun)];

  smartmouse_down(sc);

  return IFD_SUCCESS;
}

RESPONSECODE PowerUp(DWORD Lun, PUCHAR pcbuffer, PDWORD pwlength) {

  struct sc_desc_s *sc = Readers[LunToReaderIndex(Lun)];

  if (smartmouse_up(sc, pcbuffer, pwlength) != 0)
    return IFD_COMMUNICATION_ERROR;

  return IFD_SUCCESS;
}

RESPONSECODE Reset(DWORD Lun, PUCHAR pcbuffer, PDWORD pwlength) {

  struct sc_desc_s *sc = Readers[LunToReaderIndex(Lun)];

  if (smartmouse_reset(sc, pcbuffer, pwlength) != 0)
    return IFD_COMMUNICATION_ERROR;

  return IFD_SUCCESS;
}

RESPONSECODE SendCommandT0(DWORD Lun,
                           PUCHAR TxBuffer, DWORD TxLength, 
                           PUCHAR RxBuffer, PDWORD RxLength) {

  struct sc_desc_s *sc = Readers[LunToReaderIndex(Lun)];

  // APDU Case 1
  if (TxLength == 4)
    return T0Case1(sc, TxBuffer, TxLength, RxBuffer, RxLength);

  // APDU Case 2
  if (TxLength == 5) 
    return T0Case2(sc, TxBuffer, TxLength, RxBuffer, RxLength);

  
  // APDU Case 3
  if(TxBuffer[4] == (TxLength - 5)) 
    return T0Case3(sc, TxBuffer, TxLength, RxBuffer, RxLength);


  // APDU Case 4
  if(TxBuffer[4] == (TxLength - 6))
    return T0Case4(sc, TxBuffer, TxLength, RxBuffer, RxLength);
  
  return IFD_COMMUNICATION_ERROR;
}

RESPONSECODE T0Case1(struct sc_desc_s *sc,
                     PUCHAR TxBuffer, DWORD TxLength, 
                     PUCHAR RxBuffer, PDWORD RxLength) {
  DWORD dwlengthOut;
  UCHAR pcbufferHeader[5];
  UCHAR pcbufferOut[RESP_BUF_SIZE];
  RESPONSECODE rv;
  dwlengthOut = 0;

#ifdef PCSC_DEBUG
  printf("APDU Case 1\n");
#endif
  memcpy(pcbufferHeader, TxBuffer, TxLength);
  pcbufferHeader[TxLength] = 0x00;
    
  rv = SendT0Header(sc, pcbufferHeader);
  if (rv != IFD_SUCCESS)
    return rv;
    
  rv = ReceiveProceduralBytes(sc, pcbufferHeader[1], pcbufferOut, &dwlengthOut);
  if (rv != IFD_SUCCESS)
    return rv;

  rv = ReceiveSW2(sc, &pcbufferOut[dwlengthOut]);
  if (rv != IFD_SUCCESS)
    return rv;
  dwlengthOut++;

  if(dwlengthOut != 2)
    return IFD_COMMUNICATION_ERROR;
  
  //SW
  memcpy(RxBuffer, pcbufferOut, dwlengthOut);
  *RxLength = dwlengthOut;
  
  return IFD_SUCCESS;
}

RESPONSECODE T0Case2(struct sc_desc_s *sc,
                     PUCHAR TxBuffer, DWORD TxLength, 
                     PUCHAR RxBuffer, PDWORD RxLength) {

  DWORD dwlengthOut;
  UCHAR pcbufferHeader[5];
  UCHAR pcbufferOut[RESP_BUF_SIZE];
  RESPONSECODE rv;
  dwlengthOut = 0;

#ifdef PCSC_DEBUG
  printf("APDU Case 2\n");
#endif
  rv = SendT0Header(sc, TxBuffer);
  if (rv != IFD_SUCCESS)
    return rv;
  
  rv = ReceiveProceduralBytes(sc, TxBuffer[1], pcbufferOut, &dwlengthOut);
  if (rv != IFD_SUCCESS)
    return rv;

  if (TxBuffer[1] == pcbufferOut[0]) {
    rv = ReceiveData(sc, pcbufferOut, TxBuffer[4]);
    if (rv != IFD_SUCCESS)
      return rv;
    dwlengthOut += TxBuffer[4];
    
    rv = ReceiveSW1(sc, &pcbufferOut[dwlengthOut]);
    if (rv != IFD_SUCCESS)
      return rv;
    dwlengthOut++;
  }

  rv = ReceiveSW2(sc, &pcbufferOut[dwlengthOut]);
  if (rv != IFD_SUCCESS)
    return rv;
  dwlengthOut++;

  // if SW1 = 6C
  if ((dwlengthOut == 2) && (pcbufferOut[0] == 0x6C)) {
    memcpy(pcbufferHeader, TxBuffer, TxLength);
    pcbufferHeader[4] = pcbufferOut[1];

    rv = SendT0Header(sc, pcbufferHeader);
    if (rv != IFD_SUCCESS)
      return rv;

    rv = ReceiveProceduralBytes(sc, pcbufferHeader[1], pcbufferOut, &dwlengthOut);
    if (rv != IFD_SUCCESS)
      return rv;
     
    rv = ReceiveData(sc, pcbufferOut, pcbufferHeader[4]);
    if (rv != IFD_SUCCESS)
      return rv;
    dwlengthOut += pcbufferHeader[4];

    rv = ReceiveSW1(sc, &pcbufferOut[dwlengthOut]);
    if (rv != IFD_SUCCESS)
      return rv;
    dwlengthOut++;

    rv = ReceiveSW2(sc, &pcbufferOut[dwlengthOut]);
    if (rv != IFD_SUCCESS)
      return rv;
    dwlengthOut++;
 
    if (pcbufferHeader[4] > TxBuffer[4]) {
      if (TxBuffer[4] == 0) {
        memcpy(RxBuffer, pcbufferOut, dwlengthOut);
        *RxLength = dwlengthOut;
       
         return IFD_SUCCESS;  
      }

      memcpy(RxBuffer, pcbufferOut, TxBuffer[4]);
      //SW
      memcpy(&RxBuffer[TxBuffer[4]], &pcbufferOut[dwlengthOut-2], 2*sizeof(UCHAR));
      *RxLength = TxBuffer[4]+2;
        
      return IFD_SUCCESS;
    }
  }

  //SW
  memcpy(RxBuffer, pcbufferOut, dwlengthOut);
  *RxLength = dwlengthOut;

  return IFD_SUCCESS;
}

RESPONSECODE T0Case3(struct sc_desc_s *sc,
                     PUCHAR TxBuffer, DWORD TxLength, 
                     PUCHAR RxBuffer, PDWORD RxLength) {

  DWORD dwlengthOut;
  UCHAR pcbufferOut[RESP_BUF_SIZE];
  RESPONSECODE rv;
  dwlengthOut = 0;

#ifdef PCSC_DEBUG
  printf("APDU Case 3\n");
#endif
  
  rv = SendT0Header(sc, TxBuffer);
  if (rv != IFD_SUCCESS)
    return rv;
  
  rv = ReceiveProceduralBytes(sc, TxBuffer[1], pcbufferOut, &dwlengthOut);
  if (rv != IFD_SUCCESS)
    return rv;

  rv = SendT0Data(sc, &TxBuffer[5], TxLength - 5);
  if (rv != IFD_SUCCESS)
    return rv;

  rv = ReceiveProceduralBytes(sc, TxBuffer[1], pcbufferOut, &dwlengthOut);
  if (rv != IFD_SUCCESS)
    return rv;

  rv = ReceiveSW2(sc, &pcbufferOut[dwlengthOut]);
  if (rv != IFD_SUCCESS)
    return rv;
  dwlengthOut++;

  if(dwlengthOut != 2)
    return IFD_COMMUNICATION_ERROR;
    
  //SW
  memcpy(RxBuffer, pcbufferOut, dwlengthOut);
  *RxLength = dwlengthOut;

  return IFD_SUCCESS;
}

RESPONSECODE T0Case4(struct sc_desc_s *sc,
                     PUCHAR TxBuffer, DWORD TxLength, 
                     PUCHAR RxBuffer, PDWORD RxLength) {

  DWORD dwlengthOut;
  UCHAR pcbufferOut[RESP_BUF_SIZE];
  UCHAR GetResponse[] = {0x00, 0xC0, 0x00, 0x00, 0x00};
  RESPONSECODE rv;
  dwlengthOut = 0;

#ifdef PCSC_DEBUG
  printf("APDU Case 4\n");
#endif
  
  rv = T0Case3(sc, TxBuffer, TxLength-1, pcbufferOut, &dwlengthOut);
  if (rv != IFD_SUCCESS)
    return rv;

  if(TxBuffer[TxLength-1] != 0x00)
    if ((pcbufferOut[0] == 0x90) && (pcbufferOut[1] == 0x00)) {
      GetResponse[4] = TxBuffer[TxLength-1];
      return T0Case2(sc, GetResponse, sizeof(GetResponse), RxBuffer, RxLength);
    }
  
  if (pcbufferOut[0] == 0x61) {
    
#ifdef PCSC_DEBUG
    printf("SW2 = %02X\n", pcbufferOut[1]);
#endif
    if (TxBuffer[TxLength-1] > pcbufferOut[1])
      GetResponse[4] = pcbufferOut[1];
    else {
      if (TxBuffer[TxLength-1] == 0)
        GetResponse[4] = pcbufferOut[1];
      else
        GetResponse[4] = TxBuffer[TxLength-1];
    }
    
    rv = SendT0Header(sc, GetResponse);
    if (rv != IFD_SUCCESS)
      return rv;
    
    rv = ReceiveProceduralBytes(sc, GetResponse[1], pcbufferOut, &dwlengthOut);
    if (rv != IFD_SUCCESS)
      return rv;

    if (GetResponse[1] == pcbufferOut[0]) {
      rv = ReceiveData(sc, pcbufferOut, GetResponse[4]);
      if (rv != IFD_SUCCESS)
        return rv;
      dwlengthOut += GetResponse[4];
    }

    rv = ReceiveSW1(sc, &pcbufferOut[dwlengthOut]);
    if (rv != IFD_SUCCESS)
      return rv;
    dwlengthOut++;
    
    rv = ReceiveSW2(sc, &pcbufferOut[dwlengthOut]);
    if (rv != IFD_SUCCESS)
      return rv;
    dwlengthOut++;
  }

  //SW
  memcpy(RxBuffer, pcbufferOut, dwlengthOut);
  *RxLength = dwlengthOut;
  
  return IFD_SUCCESS;
}

RESPONSECODE SendT0Header(struct sc_desc_s *sc, PUCHAR TxBuffer) {
#ifdef PCSC_DEBUG
  {
    int i;
    printf("Send TPDU header: ");
    for(i = 0; i != 5; i++) 
      printf("%02X ", TxBuffer[i]);
    printf("\n");
  }
#endif
  
  if (smartmouse_write(sc, TxBuffer, 5*sizeof(UCHAR)) == -1)
    return IFD_COMMUNICATION_ERROR;
  return IFD_SUCCESS;
}


RESPONSECODE SendT0Data(struct sc_desc_s *sc,
                        PUCHAR TxBuffer, DWORD dwLength) {

#ifdef PCSC_DEBUG
  {
    int i;
    printf("Send TPDU Data: ");
    for(i = 0; i != dwLength; i++) 
      printf("%02X ", TxBuffer[i]);
    printf("\n");
  }
#endif
  
  if (smartmouse_write(sc, TxBuffer, dwLength*sizeof(UCHAR)) == -1)
    return IFD_COMMUNICATION_ERROR;
  return IFD_SUCCESS;
}


RESPONSECODE ReceiveProceduralBytes(struct sc_desc_s *sc, UCHAR Ins,
                                    PUCHAR RxBuffer, PDWORD pwLength) {
  int bytes_read = 0;
  RxBuffer[0] = 0;
  
  while((bytes_read == 0) || ((bytes_read == 1) && (RxBuffer[0] == 0x60))) {
    if ((bytes_read = smartmouse_read(sc, RxBuffer, sizeof(UCHAR))) == -1)
      return IFD_COMMUNICATION_ERROR;

#ifdef PCSC_DEBUG
    if((bytes_read == 1) && RxBuffer[0] == 0x60)
      printf("WaitExtension\n");
#endif
  }

  if (bytes_read != 1)
    return IFD_COMMUNICATION_ERROR;

  *pwLength = 0;

  if (RxBuffer[0] == Ins) {
#ifdef PCSC_DEBUG
    printf("ACK\n");
#endif
    return IFD_SUCCESS;
  }

  *pwLength = bytes_read;
  
  if((RxBuffer[0] & 0xF0) == 0x60) {
#ifdef PCSC_DEBUG
    printf("SW1 = 0x6X (%02X)\n", RxBuffer[0]);
#endif
    return IFD_SUCCESS;
  }
  
  if((RxBuffer[0] & 0xF0) == 0x90) {
#ifdef PCSC_DEBUG
    printf("SW1 = 0x9X (%02X)\n", RxBuffer[0]);
#endif
    return IFD_SUCCESS;
  }

  *pwLength = 0;
      
  if ( ((RxBuffer[0] & 0xF0) != 0x60) || ((RxBuffer[0] & 0xF0) != 0x90) ) {
#ifdef PCSC_DEBUG
    printf("Procedural byte error\n");
#endif
    return IFD_COMMUNICATION_ERROR;
  }
    
  return IFD_COMMUNICATION_ERROR;
}

RESPONSECODE ReceiveData(struct sc_desc_s *sc,
                         PUCHAR RxBuffer, DWORD dwLength) {
  DWORD dwlengthOut = 0;
  
  while (dwlengthOut < dwLength) {
    if ((dwlengthOut += smartmouse_read(sc, &RxBuffer[dwlengthOut], dwLength - dwlengthOut)) == -1)
      return IFD_COMMUNICATION_ERROR;
  }

#ifdef PCSC_DEBUG
  {
    int i;
    printf("Received TPDU Data: ");
    for(i = 0; i != dwLength; i++) 
      printf("%02X ", RxBuffer[i]);
    printf("\n");
  }
#endif

  return IFD_SUCCESS;
}


RESPONSECODE ReceiveSW1(struct sc_desc_s *sc, PUCHAR RxBuffer) {
  DWORD dwlengthOut = 0;
  
  while (dwlengthOut < 1) {
    if ((dwlengthOut += smartmouse_read(sc, &RxBuffer[dwlengthOut], sizeof(UCHAR))) == -1)
      return IFD_COMMUNICATION_ERROR;
  }

#ifdef PCSC_DEBUG
  printf("Received SW1: %02X\n", RxBuffer[0]);
#endif

  return IFD_SUCCESS;
}


RESPONSECODE ReceiveSW2(struct sc_desc_s *sc, PUCHAR RxBuffer) {
  DWORD dwlengthOut = 0;
  
  while (dwlengthOut < 1) {
    if ((dwlengthOut += smartmouse_read(sc, &RxBuffer[dwlengthOut], sizeof(UCHAR))) == -1)
      return IFD_COMMUNICATION_ERROR;
  }

#ifdef PCSC_DEBUG
  printf("Received SW2: %02X\n", RxBuffer[0]);
#endif

  return IFD_SUCCESS;
}
