/*****************************************************************
/    This file is part of PC/SC Lite driver for SM1.
/
/    PC/SC Lite driver for SM1 is free software; you can redistribute it and/or
/    modify it under the terms of the GNU General Public License as published by
/    the Free Software Foundation; either version 2 of the License, or
/    (at your option) any later version.
/
/    PC/SC Lite driver for SM1 is distributed in the hope that it will be
/    useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
/    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/    GNU General Public License for more details.
/
/    You should have received a copy of the GNU General Public License
/    along with PC/SC Lite driver for SM1; if not, write to the Free Software
/    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/
/    Copyright Damien Sauveron, 2003
/    contact : damien.sauveron@labri.fr
/
/    Based on code from the SCTK project by Alexandre Becoulet.
/    URL: http://freshmeat.net/projects/sctk/
/
/ File   :   smartmouse.c
/ Author :   Damien Sauveron <damien.sauveron@labri.fr>
/ Date   :   July 9, 2003
/ Purpose:   This provides reader specific low-level calls.
/            See http://www.linuxnet.com for more information.
/ License:   See file COPYING
/
******************************************************************/

#include <stdlib.h>
#include <stdint.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/poll.h>
#include <fcntl.h>

#include "ttys.h"
#include "smartmouse.h"

/* table for indirect to direct byte mode conversion */

static const uint8_t dir_conv_table[0x100] =
{
  0xff, 0x7f, 0xbf, 0x3f, 0xdf, 0x5f, 0x9f, 0x1f,
  0xef, 0x6f, 0xaf, 0x2f, 0xcf, 0x4f, 0x8f, 0xf,
  0xf7, 0x77, 0xb7, 0x37, 0xd7, 0x57, 0x97, 0x17,
  0xe7, 0x67, 0xa7, 0x27, 0xc7, 0x47, 0x87, 0x7,
  0xfb, 0x7b, 0xbb, 0x3b, 0xdb, 0x5b, 0x9b, 0x1b,
  0xeb, 0x6b, 0xab, 0x2b, 0xcb, 0x4b, 0x8b, 0xb,
  0xf3, 0x73, 0xb3, 0x33, 0xd3, 0x53, 0x93, 0x13,
  0xe3, 0x63, 0xa3, 0x23, 0xc3, 0x43, 0x83, 0x3,
  0xfd, 0x7d, 0xbd, 0x3d, 0xdd, 0x5d, 0x9d, 0x1d,
  0xed, 0x6d, 0xad, 0x2d, 0xcd, 0x4d, 0x8d, 0xd,
  0xf5, 0x75, 0xb5, 0x35, 0xd5, 0x55, 0x95, 0x15,
  0xe5, 0x65, 0xa5, 0x25, 0xc5, 0x45, 0x85, 0x5,
  0xf9, 0x79, 0xb9, 0x39, 0xd9, 0x59, 0x99, 0x19,
  0xe9, 0x69, 0xa9, 0x29, 0xc9, 0x49, 0x89, 0x9,
  0xf1, 0x71, 0xb1, 0x31, 0xd1, 0x51, 0x91, 0x11,
  0xe1, 0x61, 0xa1, 0x21, 0xc1, 0x41, 0x81, 0x1,
  0xfe, 0x7e, 0xbe, 0x3e, 0xde, 0x5e, 0x9e, 0x1e,
  0xee, 0x6e, 0xae, 0x2e, 0xce, 0x4e, 0x8e, 0xe,
  0xf6, 0x76, 0xb6, 0x36, 0xd6, 0x56, 0x96, 0x16,
  0xe6, 0x66, 0xa6, 0x26, 0xc6, 0x46, 0x86, 0x6,
  0xfa, 0x7a, 0xba, 0x3a, 0xda, 0x5a, 0x9a, 0x1a,
  0xea, 0x6a, 0xaa, 0x2a, 0xca, 0x4a, 0x8a, 0xa,
  0xf2, 0x72, 0xb2, 0x32, 0xd2, 0x52, 0x92, 0x12,
  0xe2, 0x62, 0xa2, 0x22, 0xc2, 0x42, 0x82, 0x2,
  0xfc, 0x7c, 0xbc, 0x3c, 0xdc, 0x5c, 0x9c, 0x1c,
  0xec, 0x6c, 0xac, 0x2c, 0xcc, 0x4c, 0x8c, 0xc,
  0xf4, 0x74, 0xb4, 0x34, 0xd4, 0x54, 0x94, 0x14,
  0xe4, 0x64, 0xa4, 0x24, 0xc4, 0x44, 0x84, 0x4,
  0xf8, 0x78, 0xb8, 0x38, 0xd8, 0x58, 0x98, 0x18,
  0xe8, 0x68, 0xa8, 0x28, 0xc8, 0x48, 0x88, 0x8,
  0xf0, 0x70, 0xb0, 0x30, 0xd0, 0x50, 0x90, 0x10,
  0xe0, 0x60, 0xa0, 0x20, 0xc0, 0x40, 0x80, 0x0
};


/* open a smart card device and return the descriptor */

struct sc_desc_s *smartmouse_open(char   *ttysdev) {
  struct sc_desc_s *sc;
  
  /* malloc smart card descriptor */
  if ((sc = malloc(sizeof (struct sc_desc_s)))) {
    /* open serial device */
    if ((sc->dev_fd = open(ttysdev, O_RDWR)) > 0) {
      int  res = 1;
      
      /* set ttys params */
      if (!ttys_init(sc)) {
        /* call protocol specific init */
        res = smartmouse_init(sc);
      }
         
      /* return descriptor ptr if ok */
      if (res)
        free(sc);
      else
        return (sc);
    }
    else
      free(sc);
  }
  
  return (NULL);
}

/* free a smart card descriptor */

void smartmouse_close(struct sc_desc_s *sc) {
  close(sc->dev_fd);
  free(sc);
}

/* init functions called on device open */

static int smartmouse_setctrl(struct sc_desc_s *sc, const int  ctrl)
{
  int tmp;

  if (ioctl(sc->dev_fd, TIOCMGET, &tmp) == -1)
    return (-1);
  tmp &= ~(TIOCM_RTS | TIOCM_CTS | TIOCM_DTR);
  tmp |= ctrl;
  return (ioctl(sc->dev_fd, TIOCMSET, &tmp));
}

int smartmouse_init(struct sc_desc_s  *sc) {
  
  int state;

  sc->mode = PHS_CONV_DIRECT;
  sc->timeout_delay = TIME_OUT;
  sc->error = "success";

  usleep(100000);
  return (0);
}

/* send reset signal to the smart card */

int smartmouse_reset(struct sc_desc_s *sc, unsigned char *pcbuffer, unsigned long *pwlength) {
  struct pollfd pfd;
  /* use poll to handle read timeout */
  pfd.fd = sc->dev_fd;
  pfd.events = POLLIN;

  /* flush ttys io buffers */
  tcflush(sc->dev_fd, TCIOFLUSH);
  
  /* set the reset line */
  if (smartmouse_setctrl(sc, TIOCM_CTS | TIOCM_DTR)) {
    sc->error = "smartmouse_smartmouse_error: internal error on file descriptor";
    return (-1);
  }

  /* unset the reset line */
  if (smartmouse_setctrl(sc, TIOCM_RTS | TIOCM_CTS | TIOCM_DTR)) {
    sc->error = "smartmouse_smartmouse_error: internal error on file descriptor";
    return (-1);
  }

  /* use poll to handle read timeout */
  if (poll(&pfd, 1, sc->timeout_delay / 500) < 1) {
    sc->error = "Card mute";
    return (-1);
  }

  if ((*pwlength = smartmouse_read(sc, pcbuffer, *pwlength)) <= 0) {
    sc->error = "Problem with the ATR";
    return (-1);
  }
  pcbuffer[*pwlength] = 0;

  sc->error = "success";
  return (0);
}

int smartmouse_up(struct sc_desc_s *sc, unsigned char *pcbuffer, unsigned long *pwlength) {
  struct pollfd pfd;
  /* use poll to handle read timeout */
  pfd.fd = sc->dev_fd;
  pfd.events = POLLIN;
  
  smartmouse_down(sc);
  
  /* unset the reset and DTR lines */
  if (smartmouse_setctrl(sc, TIOCM_RTS | TIOCM_CTS | TIOCM_DTR)) {
    sc->error = "smartmouse_smartmouse_error: internal error on file descriptor";
    return (-1);
  }
  
  /* use poll to handle read timeout */
  if (poll(&pfd, 1, sc->timeout_delay / 500) < 1) {
    sc->error = "Card mute";
    return (-1);
  }
  
  if (smartmouse_read(sc, pcbuffer, *pwlength) <= 0) {
    sc->error = "Problem with the ATR";
    return (-1);
  }
  
  if ((pcbuffer[0] == 0x03) || (pcbuffer[0] == 0x23)) {
    if ((pcbuffer[0] == 0x03)) 
      smartmouse_command(sc, phs_command_indir_mode);
    else
      smartmouse_command(sc, phs_command_direct_mode);
  }
  
  smartmouse_down(sc);
  
  /* unset the reset and DTR lines */
  if (smartmouse_setctrl(sc, TIOCM_RTS | TIOCM_CTS | TIOCM_DTR)) {
    sc->error = "smartmouse_smartmouse_error: internal error on file descriptor";
    return (-1);
  }
  
  /* use poll to handle read timeout */
  if (poll(&pfd, 1, sc->timeout_delay / 500) < 1) {
    sc->error = "Card mute";
    return (-1);
  }
  
  if ((*pwlength = smartmouse_read(sc, pcbuffer, *pwlength)) <= 0) {
    sc->error = "Problem with the ATR";
    return (-1);
  }
  pcbuffer[*pwlength] = 0;
  
  sc->error = "success";
  return (0);
}

int smartmouse_down(struct sc_desc_s *sc) {
  /* flush ttys io buffers */
  tcflush(sc->dev_fd, TCIOFLUSH);
  
  /* set the reset and DTR lines */
  if (smartmouse_setctrl(sc, TIOCM_CTS)) {
    sc->error = "smartmouse_smartmouse_error: internal error on file descriptor";
    return (-1);
  }
  
  sc->error = "success";
  return (0);
}


int smartmouse_ispresent(struct sc_desc_s *sc) {
  int state;
  
  /* flush ttys io buffers */
  tcflush(sc->dev_fd, TCIOFLUSH);
  
  ioctl(sc->dev_fd, TIOCMGET, &state);
  state &= TIOCM_CTS;
  sc->error = "success";
  
  if (state == TIOCM_CTS)
    // Card absent
    return 0;
  else
    // Card present
    return 1;
}

int smartmouse_read(struct sc_desc_s *sc,
                    unsigned char *data,
                    const unsigned int size) {
  int i;
  
  /* read size bytes from ttys */
  for (i = 0; i < size; i++) {
    struct pollfd pfd;
    
    /* use poll to handle read timeout */
    pfd.fd = sc->dev_fd;
    pfd.events = POLLIN;
    
    if (poll(&pfd, 1, sc->timeout_delay / 1000) < 1) {
      sc->error = "success";
      return (i);
    }
    
    if (read(sc->dev_fd, &data[i], sizeof(unsigned char)) <= 0) {
      sc->error = "smartmouse_smartmouse_error: can't read data from card";
      return (-1);
    }
    
    if (sc->mode & PHS_CONV_INDIRECT) 
      data[i] =  dir_conv_table[data[i]];
  }
  
  sc->error = "success";
  return (size);
}

static int smartmouse_write_b(struct sc_desc_s *sc,
                              const unsigned char *data,
                              const unsigned int size) {
  unsigned int  i;

  /* flush ttys io buffers */
  tcflush(sc->dev_fd, TCIOFLUSH);

  /* write char by char data */
  for (i = 0; i < size; i++) {
    if (write(sc->dev_fd, &data[i], sizeof(unsigned char)) != sizeof(unsigned char)) {
      sc->error = "smartmouse_smartmouse_error: can't write data to card";
      return (-1);
    }

    tcdrain(sc->dev_fd);
  }

  /* read back written bytes and compare */
  for (i = 0; i < size; i++) {
    struct pollfd pfd;
    unsigned char  c;
    
    /* use poll to handle read timeout */
    pfd.fd = sc->dev_fd;
    pfd.events = POLLIN;
    
    if (poll(&pfd, 1, sc->timeout_delay / 1000) < 1) {
      sc->error = "smartmouse_smartmouse_error: poll error on file descriptor";
      return (-1);
    }

    if ((read(sc->dev_fd, &c, sizeof(unsigned char)) != sizeof(unsigned char)) || (c != data[i])) {
      sc->error = "smartmouse_smartmouse_error: can't read data back from card";
      return (-1);
    }
  }
 
  sc->error = "success";
  return size;
}

int smartmouse_write(struct sc_desc_s *sc,
                     const unsigned char *data,
                     const unsigned int size) {
  if (!size) {
    sc->error = "success";
    return (0);
  }

  if (sc->mode & PHS_CONV_INDIRECT) {
    unsigned char *buffer;
    int  i;
    
    if (!(buffer = malloc(size))) {
      sc->error = "malloc error";
      return (0);
    }
    
    for (i = 0; i < size; i++)
      buffer[i] = dir_conv_table[data[i]];
    
    i = smartmouse_write_b(sc, buffer, size);
    free(buffer);
    return (i);
  }
  else
    return (smartmouse_write_b(sc, data, size));
}

int smartmouse_command(struct sc_desc_s *sc,
                       const int  command) {
#ifdef SMARTMOUSE_DEBUG
  printf("SMARTMOUSE_COMMAND: ");
#endif

  switch (command) {
  case (phs_command_direct_mode):
#ifdef SMARTMOUSE_DEBUG
    printf("DIRECT\n");
#endif
    sc->mode = PHS_CONV_DIRECT;
    sc->ios.c_cflag &= ~PARODD;
    break;
    
  case (phs_command_indir_mode):
#ifdef SMARTMOUSE_DEBUG
    printf("INDIRECT\n");
#endif
    sc->mode = PHS_CONV_INDIRECT;
    sc->ios.c_cflag |= PARODD;
    break;
    
  default:
    sc->error = "smartmouse_smartmouse_error: unknown smartmouse/smartmouse command";
    return (-1);
  }
  
  if (tcsetattr(sc->dev_fd, TCSANOW, &sc->ios)) {
    sc->error = "smartmouse_smartmouse_error: internal error on file descriptor";
    return (-1);
  }

  sc->error = "success";
  return (0);
}

int smartmouse_ispending(struct sc_desc_s *sc) {
  struct pollfd  pfd;
  
  pfd.fd = sc->dev_fd;
  pfd.events = POLLIN;
  
  if (poll(&pfd, 1, 0) < 0) {
    sc->error = "smartmouse_smartmouse_error: poll error on file descriptor";
    return (-1);
  }

  sc->error = "success";
  return (pfd.revents & POLLIN ? 1 : 0);
}

